#structures.py
import asyncio
import inspect
from .exceptions import UncancellableEvent, UnfulfilledRequirement


# Wrapper for event handlers
class EventHandler:
    def __init__(self, function, event: str, priority: int, event_space, additional_kwargs: dict = {}, execution_requirements: dict = {}, requirement_exceptions: bool = False, is_async: bool = False):
        self.function = function
        self.event = event
        self.priority = priority
        self.additional_kwargs = additional_kwargs
        self.event_space = event_space
        self.execution_requirements = execution_requirements
        self.requirement_exceptions = requirement_exceptions
        self.is_async = is_async

    @property
    def origin_module(self):
        return inspect.getmodule(self.function)

    def register(self, additional_kwargs: dict={}):
        self.event_space.register_handler(self)        
        self.additional_kwargs = self.additional_kwargs | additional_kwargs

    def unregister(self):
        self.event_space.unregister_handler(self)        

    @property
    def registered(self):
        return self.event_space.handler_registered(self)

    async def async_call(self, call, args, kwargs):
        kwargs = kwargs | self.additional_kwargs
        for argument in self.execution_requirements:
            if argument not in kwargs or self.execution_requirements[argument] != kwargs[argument]:
                if self.requirement_exceptions:
                    raise UnfulfilledRequirement(f"Argument {argument} was unmet.")
                return
        return await self.function(event=call, *args, **kwargs)


    def call(self, call, args, kwargs):        
        kwargs = kwargs | self.additional_kwargs
        for argument in self.execution_requirements:
            if argument not in kwargs or self.execution_requirements[argument] != kwargs[argument]:
                if self.requirement_exceptions:
                    raise UnfulfilledRequirement(f"Argument {argument} was unmet.")
                return
        if self.is_async:
            return asyncio.run(self.function(event=call, *args, **kwargs))
        return self.function(event=call, *args, **kwargs)




# Add support for event calls inside objects
class EventSupport:
    def __init__(self):
        self.register_events()

    def register_events(self):
        for method in self.__dir__():
            method = self.__getattribute__(method)
            if isinstance(method, EventHandler):
                if method.registered:
                    method.unregister()
                method.register({"self": self})



# Define the EventCall class
class EventCall:
    def __init__(self, event_name, cancellable=False):
        self.__name = event_name
        self.__cancelled = False
        self.__stopped = False
        self.__response = None
        self.cancellable = cancellable        
        self.responses = []
    
    # Cancelling is only one-way. 
    def cancel(self):
        if self.cancellable:
            self.__cancelled = True
            return
        raise UncancellableEvent(f'{self.event_name} is not an cancellable event.')
    
    def stop(self):
        self.__stopped = True

    # Using it as a property since we don't want it to be writable 
    @property
    def cancelled(self):
        return self.__cancelled

    @property
    def response(self):
        return next((i for i in self.responses if i), None)

    @property
    def stopped(self):
        return self.__stopped

    @property
    def name(self):
        return self.__name
    
    def __repr__(self):
        return f'<EventCall _name={self.name!r} _cancelled={self.cancelled!r} _response={self.response!r}>'


from .methods import on, call, async_call, handlers_registered, register_handler, unregister_from_module, unregister_handler
from .structures import EventCall, EventHandler, EventSupport
from .exceptions import UncancellableEvent
from .event_space import EventSpace

__name__ = "pyding"
__version__ = "1.6.0.1"


import asyncio
from .structures import EventHandler, EventCall

class EventSpace:
    """
        Manages and holds the event listener/handlers.
    """
    def __init__(self):
        self.events = {}
    
    def register_handler(self, handler):
        if handler.event not in self.events:
            self.events[handler.event]: dict[str, dict[int, list[EventHandler]]] = {}
        
        # Check if the priority is already registered.
        if handler.priority not in self.events[handler.event]:
            self.events[handler.event][handler.priority] = []

        self.events[handler.event][handler.priority].append(handler)


    def get_handlers(self, event):
        # Get the ordered priorities
        index_order = list(self.events[event])
        index_order.sort(reverse=True) # Reverse the order to make it descending.


        # Index the handlers.
        handlers = []
        for index in index_order:
            for handler in self.events[event][index]:
                handlers.append(handler)
        return handlers


    def handler_registered(self, handler):
        return handler in self.events[handler.event][handler.priority] if handler.event in self.events else False

    def unregister_from_module(self, module):
        for event in self.events.values():
            for handlers in event.values():
                for handler in handlers:
                    if handler.origin_module == module:
                        handlers.remove(handler)

    def unregister_handler(self, handler):
        self.events[handler.event][handler.priority].remove(handler)


    # Define the "on" method
    def on(self, event_name: str, priority: int=0, register_ra: bool=True, function: bool=None, requirement_exceptions: bool=False, is_async: bool=False, **kwargs):
        """Attaches a handler

        Args:
            event_name (str): Event name
            priority (int, optional): Handler priority. Defaults to 0.
            register_ra (bool, optional): Register the handler rigth now. Defaults to True.
            function (bool, optional): Handler function

        Returns:
            EventHandler
        """
        # Wrap the function
        def wrapper(func):
            # Insert the handler function into the dict.
            handler = EventHandler(func, event_name, priority, self, is_async=is_async, requirement_exceptions=requirement_exceptions, execution_requirements=kwargs)
            if register_ra:
                handler.register()
            return handler
        
        # If there isn't a function in the args, return the wrapper. 
        if not function:
            return wrapper
        
        # Run the wrapper function with the argument's function.
        wrapper(function)


    async def async_call(self, event_name: str, cancellable: bool=False, blocking: bool=True, first_response: bool=False, *args, **kwargs):
        """Calls an event with asynchronous handlers

        Args:
            event_name (str): Name of the event being called.
            cancellable (bool, optional): If the called event can be cancelled. Defaults to False.
            blocking (bool, optional): Stop the event if it has been cancelled. Defaults to True.
            first_response (bool, optional): Stop the event at the first response. Defaults to False.

        Returns:
            EventCall
        """
        # Define the event call object.
        event_call = EventCall(event_name=event_name, cancellable=cancellable)
 
        # Return if there isn't an event registered with this name.
        if event_name not in self.events:
            return event_call

        # Run the handlers.
        calls = []
        for handler in self.get_handlers(event_name):
            handler: EventHandler
            calls.append(handler.async_call(event_call, args=args, kwargs=kwargs))
        event_call.responses = await asyncio.gather(*calls)
        # Return the event
        return event_call

    # Define the "call" method
    def call(self, event_name: str, cancellable: bool=False, blocking: bool=True, first_response: bool=False, *args, **kwargs):
        """Calls an event

        Args:
            event_name (str): Name of the event being called.
            cancellable (bool, optional): If the called event can be cancelled. Defaults to False.
            blocking (bool, optional): Stop the event if it has been cancelled. Defaults to True.
            first_response (bool, optional): Stop the event at the first response. Defaults to False.

        Returns:
            EventCall
        """
        
        # Define the event call object.
        event_call = EventCall(event_name=event_name, cancellable=cancellable)

        # Return if there isn't an event registered with this name.
        if event_name not in self.events:
            return event_call

        # Run the handlers.
        for handler in self.get_handlers(event_name):
            handler: EventHandler
            # Run the handler
            response = handler.call(event_call, args=args, kwargs=kwargs)
            event_call.responses.append(response)
            if response != None:
                # If this is the first response, break the loop
                if first_response:
                    break

            # If the event was cancelled, do not run the next handlers.
            if (event_call.cancelled and blocking) or event_call.stopped:
                break

        # Return the event
        return event_call

global_event_space = EventSpace()


# methods.py
from .event_space import global_event_space
from .structures import EventCall, EventHandler


on = global_event_space.on
call = global_event_space.call
async_call = global_event_space.async_call
handlers_registered = global_event_space.handler_registered
register_handler = global_event_space.register_handler
unregister_from_module = global_event_space.unregister_from_module
unregister_handler = global_event_space.unregister_handler
get_handlers = global_event_space.get_handlers



# exceptions.py 

# Define the exception for cancelling an uncancellable event.
class UncancellableEvent(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class UnfulfilledRequirement(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)